@@ -21,18 +21,18 @@
 
 /**
  * SECTION:element-midiparse
- * @see_also: fluidsynth
+ * @title: midiparse
+ * @see_also: fluiddec
  *
  * This element parses midi-files into midi events. You would need a midi
  * renderer such as fluidsynth to convert the events into raw samples.
  *
- * <refsect2>
- * <title>Example pipeline</title>
+ * ## Example pipeline
  * |[
- * gst-launch-1.0 filesrc location=song.mid ! midiparse ! fluidsynth ! pulsesink
+ * gst-launch-1.0 filesrc location=song.mid ! midiparse ! fluiddec ! pulsesink
  * ]| This example pipeline will parse the midi and render to raw audio which is
  * played via pulseaudio.
- * </refsect2>
+ *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -118,6 +118,8 @@ static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
 
 #define parent_class gst_midi_parse_parent_class
 G_DEFINE_TYPE (GstMidiParse, gst_midi_parse, GST_TYPE_ELEMENT);
+GST_ELEMENT_REGISTER_DEFINE (midiparse, "midiparse", GST_RANK_PRIMARY,
+    GST_TYPE_MIDI_PARSE);
 
 /* initialize the plugin's class */
 static void
@@ -133,10 +135,8 @@ gst_midi_parse_class_init (GstMidiParseClass * klass)
   gobject_class->set_property = gst_midi_parse_set_property;
   gobject_class->get_property = gst_midi_parse_get_property;
 
-  gst_element_class_add_pad_template (gstelement_class,
-      gst_static_pad_template_get (&src_factory));
-  gst_element_class_add_pad_template (gstelement_class,
-      gst_static_pad_template_get (&sink_factory));
+  gst_element_class_add_static_pad_template (gstelement_class, &src_factory);
+  gst_element_class_add_static_pad_template (gstelement_class, &sink_factory);
   gst_element_class_set_static_metadata (gstelement_class, "MidiParse",
       "Codec/Demuxer/Audio",
       "Midi Parser Element", "Wim Taymans <wim.taymans@gmail.com>");
@@ -829,7 +829,7 @@ parse_MTrk (GstMidiParse * midiparse, guint8 * data, guint size)
   if (midiparse->track_count >= midiparse->ntracks)
     return TRUE;
 
-  track = g_slice_new (GstMidiTrack);
+  track = g_new (GstMidiTrack, 1);
   track->data = data;
   track->size = size;
   reset_track (track, midiparse);
@@ -1284,9 +1284,7 @@ pause:
       event = gst_event_new_eos ();
       /* for fatal errors we post an error message, post the error
        * first so the app knows about the error first. */
-      GST_ELEMENT_ERROR (midiparse, STREAM, FAILED,
-          ("Internal data flow error."),
-          ("streaming task paused, reason %s (%d)", reason, ret));
+      GST_ELEMENT_FLOW_ERROR (midiparse, ret);
       gst_pad_push_event (midiparse->srcpad, event);
     }
   }
@@ -1295,7 +1293,7 @@ pause:
 static void
 free_track (GstMidiTrack * track, GstMidiParse * midiparse)
 {
-  g_slice_free (GstMidiTrack, track);
+  g_free (track);
 }
 
 static void
