@@ -83,7 +83,6 @@ struct _GstD3D11WindowWin32
   HWND external_hwnd;
   GstD3D11WindowWin32OverlayState overlay_state;
 
-  HDC device_handle;
   gboolean have_swapchain1;
 
   /* atomic */
@@ -354,16 +353,10 @@ gst_d3d11_window_win32_set_render_rectangle (GstD3D11Window * window,
        * on message pumping thread is not a worst idea in generall */
       PostMessageA (self->internal_hwnd, WM_GST_D3D11_MOVE_WINDOW, 0, 0);
     }
-  } else {
-    /* XXX: Not sure what's expected behavior if we are drawing on internal
-     * HWND but user wants to specify rectangle.
-     *
-     * - Should we move window to corresponding desktop coordinates ?
-     * - Or should crop correspondingly by modifying viewport of
-     *   render target view of swapchian's backbuffer or so ?
-     * - Or should we ignore set_render_rectangle if we are drawing on
-     *   internal HWND without external HWND ?
-     */
+  } else if (!window->external_handle && self->internal_hwnd) {
+    MoveWindow (self->internal_hwnd,
+        self->render_rect.x, self->render_rect.y, self->render_rect.w,
+        self->render_rect.h, TRUE);
   }
 }
 
@@ -595,7 +588,7 @@ gst_d3d11_window_win32_create_internal_window (GstD3D11WindowWin32 * self)
     wc.lpfnWndProc = window_proc;
     wc.hInstance = hinstance;
     wc.hIcon = LoadIcon (NULL, IDI_WINLOGO);
-    wc.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
+    wc.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
     wc.hCursor = LoadCursor (NULL, IDC_ARROW);
     wc.hbrBackground = (HBRUSH) GetStockObject (BLACK_BRUSH);
     wc.lpszClassName = "GSTD3D11";
@@ -612,7 +605,6 @@ gst_d3d11_window_win32_create_internal_window (GstD3D11WindowWin32 * self)
     GST_LOG_OBJECT (self, "window class was already registered");
   }
 
-  self->device_handle = 0;
   self->internal_hwnd = 0;
   self->visible = FALSE;
 
@@ -632,12 +624,6 @@ gst_d3d11_window_win32_create_internal_window (GstD3D11WindowWin32 * self)
   GST_DEBUG_OBJECT (self, "d3d11 window created: %" G_GUINTPTR_FORMAT,
       (guintptr) self->internal_hwnd);
 
-  /* device_handle is set in the window_proc */
-  if (!self->device_handle) {
-    GST_ERROR_OBJECT (self, "device handle is not available");
-    return FALSE;
-  }
-
   GST_LOG_OBJECT (self,
       "Created a internal d3d11 window %p", self->internal_hwnd);
 
@@ -880,12 +866,6 @@ window_proc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
 
     GST_LOG_OBJECT (self, "WM_CREATE");
 
-    self->device_handle = GetDC (hWnd);
-    /* Do this, otherwise we hang on exit. We can still use it (due to the
-     * CS_OWNDC flag in the WindowClass) after we have Released.
-     */
-    ReleaseDC (hWnd, self->device_handle);
-
     SetPropA (hWnd, D3D11_WINDOW_PROP_NAME, self);
   } else if ((self = gst_d3d11_window_win32_hwnd_get_instance (hWnd))) {
     g_rec_mutex_lock (&self->hwnds_lock);
@@ -1215,22 +1195,30 @@ gst_d3d11_window_win32_show (GstD3D11Window * window)
     /* if no parent the real size has to be set now because this has not been done
      * when at window creation */
     if (!self->external_hwnd) {
-      RECT rect = { 0, };
-
-      rect.right = width;
-      rect.bottom = height;
-
-      if (AdjustWindowRect (&rect, WS_GST_D3D11, FALSE)) {
-        width = rect.right - rect.left;
-        height = rect.bottom - rect.top;
+      if (self->render_rect.x != 0 || self->render_rect.y != 0 ||
+          self->render_rect.w != 0 || self->render_rect.h != 0) {
+        MoveWindow (self->internal_hwnd,
+            self->render_rect.x, self->render_rect.y, self->render_rect.w,
+            self->render_rect.h, FALSE);
       } else {
-        width += 2 * GetSystemMetrics (SM_CXSIZEFRAME);
-        height +=
-            2 * GetSystemMetrics (SM_CYSIZEFRAME) +
-            GetSystemMetrics (SM_CYCAPTION);
+        RECT rect = { 0, };
+
+        rect.right = width;
+        rect.bottom = height;
+
+        if (AdjustWindowRect (&rect, WS_GST_D3D11, FALSE)) {
+          width = rect.right - rect.left;
+          height = rect.bottom - rect.top;
+        } else {
+          width += 2 * GetSystemMetrics (SM_CXSIZEFRAME);
+          height +=
+              2 * GetSystemMetrics (SM_CYSIZEFRAME) +
+              GetSystemMetrics (SM_CYCAPTION);
+        }
+
+        MoveWindow (self->internal_hwnd, 0, 0, width, height, FALSE);
       }
 
-      MoveWindow (self->internal_hwnd, 0, 0, width, height, FALSE);
       ShowWindow (self->internal_hwnd, SW_SHOW);
     } else if (self->internal_hwnd) {
       /* ShowWindow will throw message to message pumping thread (app thread)
@@ -1315,6 +1303,13 @@ gst_d3d11_window_win32_change_fullscreen_mode (GstD3D11Window * window)
   }
 }
 
+HWND
+gst_d3d11_window_win32_get_internal_hwnd (GstD3D11Window * window)
+{
+  GstD3D11WindowWin32 *self = GST_D3D11_WINDOW_WIN32 (window);
+  return self->internal_hwnd;
+}
+
 GstD3D11Window *
 gst_d3d11_window_win32_new (GstD3D11Device * device, guintptr handle)
 {
