@@ -197,15 +197,14 @@ gst_mf_device_provider_device_enum_completed (GstWinRTDeviceWatcher *
     watcher, gpointer user_data);
 #endif
 
+static void
+gst_mf_device_provider_on_device_updated (GstMFDeviceProvider * self);
+
 static void
 gst_mf_device_provider_class_init (GstMFDeviceProviderClass * klass)
 {
-  auto object_class = G_OBJECT_CLASS (klass);
   GstDeviceProviderClass *provider_class = GST_DEVICE_PROVIDER_CLASS (klass);
 
-  object_class->dispose = gst_mf_device_provider_dispose;
-  object_class->finalize = gst_mf_device_provider_finalize;
-
   provider_class->probe = GST_DEBUG_FUNCPTR (gst_mf_device_provider_probe);
   provider_class->start = GST_DEBUG_FUNCPTR (gst_mf_device_provider_start);
   provider_class->stop = GST_DEBUG_FUNCPTR (gst_mf_device_provider_stop);
